#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>

#include "taxml.h"

#define INDENT_STRING "    "

int indent_level = 0;

int opt_positivate = 0;
int opt_dualize    = 0;

#define VATA_OUTPUT  0
#define DUMP_OUTPUT  -1

int opt_output = DUMP_OUTPUT;

char *file_name = NULL;


Ta mk_ta(int vars_no, char **var_list, int state_no, int init_state, int acc_state) {
  Ta ta = (Ta) malloc (sizeof(struct ta));

  ta->vars_no = vars_no;
  ta->var_type = (int *) malloc (vars_no * sizeof(int));

  int i;

  for (i = 0; i < vars_no; i ++) {
    char c = *(var_list[i]);
    
    if (islower(c))
      ta->var_type[i] = FOVAR;
    else if (c == UNIVAR_SYM)
      ta->var_type[i] = UNIVAR;
    else if (isupper(c))
      ta->var_type[i] = MSOVAR;
    else {
      printf("illegal variable name: %s\n", var_list[i]);
      return NULL;
    }
  }

  ta->var_list = var_list;
  ta->state_no = state_no;
  ta->init_state = init_state;
  ta->acc_state = acc_state;

  ta->trans_no = 0;
  ta->trans_list = (Trans *) malloc (MAXTRANS * sizeof(struct trans));

  return ta;
}


void ta_add_trans(Ta ta, int src0, int src1, char *lab, int dst) {
  Trans trans = (Trans) malloc (sizeof(struct trans));

  trans->src0 = src0;
  trans->src1 = src1;
  
  trans->lab = (char *) malloc (MAXVAR);
  strcpy(trans->lab, lab);

  trans->dst = dst;
  
  ta->trans_list[ta->trans_no ++] = trans;
}


void ta_print_dump(Ta ta) {
  int i;

  printf("alphabet: ");
  
  for (i = 0; i < ta->vars_no; i ++)
    printf("%s (%d)%s", ta->var_list[i], ta->var_type[i], i < ta->vars_no - 1 ? ", " : "\n");

  printf("%d states\n", ta->state_no);
  printf("initial state: %d\n", ta->init_state);
  printf("accepting state: %d\n", ta->acc_state);

  for (i = 0; i < ta->trans_no; i ++) {
    Trans trans = ta->trans_list[i];
    printf("%s(%d, %d) -> %d\n", trans->lab, trans->src0, trans->src1, trans->dst);  
  }
}


void ta_pos(Ta ta) {
  int i;

  for (i = 0; i < ta->trans_no; i ++) {
    Trans trans = ta->trans_list[i];

    int j;

    for (j = 0; j < ta->vars_no; j ++) {
      if (ta->var_type[j] == MSOVAR &&
	  trans->lab[j] == '0')
	trans->lab[j] = 'X';
    }
  }
}


void ta_dual(Ta ta) {
  int i;

  for (i = 0; i < ta->trans_no; i ++) {
    Trans trans = ta->trans_list[i];

    int j;

    for (j = 0; j < ta->vars_no; j ++) {
      if (ta->var_type[j] == MSOVAR &&
	  trans->lab[j] == '1')
	trans->lab[j] = '0';
    }
  }
}


void output_noindent(char *fmt, ...) {
  va_list args;
  
  va_start(args, fmt);
  vprintf(fmt, args);
  va_end(args);
}


void output(char *fmt, ...) {
  va_list args;
  int i;
  
  for(i = 0; i < indent_level; i ++)
    printf(INDENT_STRING);
  
  va_start(args, fmt);
  vprintf(fmt, args);
  va_end(args);
}


void vata_prologue() {
  output("Automaton ");

  if (file_name)
    output("%s\n", file_name);
  else
    output("AAA\n");
}


void vata_states(int state_no) {
  output("States");

  int i;

  for (i = 0; i < state_no; i ++)
    output(" q%d", i);

  output("\n");
}


void vata_accepting(int acc_no) {
  output("Final States %d\n", acc_no);
}


void vata_transitions(Ta ta) {
  output("Transitions\n");

  int i;

  for (i = 0; i < ta->trans_no; i ++) {
    Trans trans = ta->trans_list[i];

    int j, k;

    char vata_lab[MAXVAR+1]; 
    
    for (j = 0; j < MAXVAR - ta->vars_no; j ++)
      vata_lab[j] = 'X';
    
    for (k = 0; k < ta->vars_no; k ++)
      vata_lab[j + k] =  trans->lab[k];

    vata_lab[j + k] = '\0';

    if (trans->src0 == ta->init_state && trans->src1 == ta->init_state)
      output("%s -> q%d\n", vata_lab, trans->dst);

    output("%s(q%d, q%d) -> q%d\n", vata_lab, trans->src0, trans->src1, trans->dst);
  }
}


void ta_print_vata(Ta ta) {
  vata_prologue();
  vata_states(ta->state_no);
  vata_accepting(ta->acc_state);
  vata_transitions(ta);
}


int parse_options(int argc, char **argv) {
  int pos = 1;

  while (pos < argc-1) {
    if (!strcmp(argv[pos], "-p"))
      opt_positivate = 1;
    else if (!strcmp(argv[pos], "-d"))
      opt_dualize = 1;
    else if (!strcmp(argv[pos], "-l")) {
      pos ++;
      
      if (!strcmp(argv[pos], "vata"))
	opt_output = VATA_OUTPUT;
      else if (!strcmp(argv[pos], "dump"))
	opt_output = DUMP_OUTPUT;
      else {
	printf("unknown output format: %s\n", argv[pos]);
	return 0;
      }
    }
    
    pos ++;
  }

  if (opt_dualize && !opt_positivate) {
    printf("option -d requires -p\n");
    return 0;
  }
  
  return pos;
}


int starts_with(char *pre, char *str) {
  int len_pre = strlen(pre);
  int len_str = strlen(str);
  
  return len_str < len_pre ? 0 : !strncmp(pre, str, len_pre);
}


char *skip_until(FILE *file, char *str) {
  char *linebuf = (char *) malloc(LINESIZE);
  linebuf[0] = '\0';
  
    while (!feof(file)) {
      
      if (!fgets(linebuf, LINESIZE, file)) {
	printf("%s: error reading file [1]\n", file_name);
	return NULL;
      }

      if (str) {
	if (starts_with(str, linebuf))
	  break;
      }
    }
    
    if (feof(file)) {
      printf("%s: unexpected end of file [1]\n", file_name);
      return NULL;
    }
    
    return linebuf;
}


int parse_trans(char *buf, int len, int *src0, int *src1, char *lab, int *dst) {
  int start_lhs, fst_comma, snd_comma, end_lhs;
  
  char *src0_buf = (char *) malloc (len);
  char *src1_buf = (char *) malloc (len);
  char *dst_buf = (char *) malloc (len);
  
  if (buf[0] != '(')
    return 0;
  
  int i = start_lhs = 1;

  while (i < len && buf[i ++] != ',');
  
  if (i == len) {
    printf("error parsing transition [1]\n");
    return 0;
  }

  strncpy(src0_buf, buf+start_lhs, i-start_lhs-1);
  sscanf(src0_buf, "%d", src0);
  
  fst_comma = i;
  
  while (i < len && buf[i ++] != ',');

  if (i == len) {
    printf("error parsing transition [2]\n");
    return 0;
  }

  strncpy(src1_buf, buf+fst_comma, i-fst_comma-1);
  sscanf(src1_buf, "%d", src1);
  
  snd_comma = i;

  while (i < len && buf[i ++] != ')');

  if (i == len) {
    printf("error parsing transition [3]\n");
    return 0;
  }

  strncpy(lab, buf+snd_comma, i-snd_comma-1);
  
  end_lhs = i += strlen(" -> ");

  while (i < len && buf[i ++] != '\n');
  
  strncpy(dst_buf, buf+end_lhs, i - end_lhs);
  sscanf(dst_buf, "%d", dst);

  return 1;
}


int main(int argc, char **argv) {
  FILE *file;

  if (argc <= 1) {
    printf("usage: taxml <options> <file>\n");
    return 0;
  }

  int file_pos = parse_options(argc, argv);

  if (file_pos <= 0)
    return 0;

  if (file_pos >= argc) {
    printf("usage: autlog <options> <file>\n");
    return 0;    
  }
  
  file_name = argv[file_pos];
  
  file = fopen(file_name, "r");
  
  if (!file) {
    printf("%s: file not found\n", file_name);
    return 0;
  }

  /* read preamble */
  
  char **var_list = (char **) malloc(MAXVAR * sizeof(char *));
  char *var = (char *) malloc(LINESIZE);
  int var_no = 0;

  char * linebuf = skip_until(file, "GTA for formula with free variables: "); 

  if (!linebuf)
    return 0;
  
  int start = strlen("DFA for formula with free variables: ");

  while (sscanf(linebuf+start, "%s ", var) == 1) {
    var_list[var_no] = (char *) malloc(strlen(var));
    strcpy(var_list[var_no ++], var);
    start += strlen(var) + 1;
  }
  
  if (!var_no) {
    printf("%s: the automaton has no variables\n", argv[1]);
    return 0;
  }

  int init_state;
  int acc_state;
  int rej_state;

  /*
   * Accepting and rejecting states are switched on purpose. 
   * MONA seems to get them the other way around. 
   */
  
  fscanf(file, "Accepting states: %d\n", &rej_state);
  fscanf(file, "Rejecting states: %d\n", &acc_state);

  if (!(linebuf = skip_until(file, "State space 1 \'<univ>\'")))
    return 0;

  int state_no;
  
  sscanf(linebuf, "State space 1 '<univ>' (size %d):\n", &state_no);
  fscanf(file, "Initial state: %d\n", &init_state);

  Ta ta = mk_ta(var_no, var_list, state_no, init_state, acc_state);
  
  /* read transitions */
  
  if (!(linebuf = skip_until(file, "Transitions:")))
    return 0;

  int src0, src1, dst;
  char *lab = (char *) malloc (MAXVAR);
 
  while(!feof(file)) {
    
    if (!fgets(linebuf, LINESIZE, file)) {
      printf("%s: error reading file [2]\n", file_name);
      return 0;
    }

    if (!parse_trans(linebuf, strlen(linebuf), &src0, &src1, lab, &dst))
      break;
    
    if (src0 < 0 || src0 >= state_no) {
      printf("left source state %d out of range (%d)\n", src0, state_no);
      return 0;
    }

    if (src1 < 0 || src1 >= state_no) {
      printf("right source state %d out of range (%d)\n", src1, state_no);
      return 0;
    }
    
    if (dst < 0 || dst >= state_no) {
      printf("destination state %d out of range (%d)\n", dst, state_no);
      return 0;
    }    
        
    ta_add_trans(ta, src0, src1, lab, dst);
  }
  
  if (feof(file)) {
    printf("%s: unexpected end of file [2]\n", file_name);
    return 0;
  }
  
  /* positivate and dualize transition labels */
  
  if (opt_positivate) {
    ta_pos(ta);

    if (opt_dualize)
      ta_dual(ta);
  }

  if (opt_output == VATA_OUTPUT)
    ta_print_vata(ta);
  else if (opt_output == DUMP_OUTPUT)
    ta_print_dump(ta);
  else {
    printf("unknown print format: \%d\n", opt_output);
    return 0;
  }

  return 1;
}
